/*********************************************************************
 *
 * AUTHORIZATION TO USE AND DISTRIBUTE
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: 
 *
 * (1) source code distributions retain this paragraph in its entirety, 
 *  
 * (2) distributions including binary code include this paragraph in
 *     its entirety in the documentation or other materials provided 
 *     with the distribution, and 
 *
 * (3) all advertising materials mentioning features or use of this 
 *     software display the following acknowledgment:
 * 
 *      "This product includes software written and developed 
 *       by Brian Adamson and Joe Macker of the Naval Research 
 *       Laboratory (NRL)." 
 *         
 *  The name of NRL, the name(s) of NRL  employee(s), or any entity
 *  of the United States Government may not be used to endorse or
 *  promote  products derived from this software, nor does the 
 *  inclusion of the NRL written and developed software  directly or
 *  indirectly suggest NRL or United States  Government endorsement
 *  of this product.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 ********************************************************************/
 
 
#ifndef _PROTO_TIMER
#define _PROTO_TIMER

#include "protoDebug.h"
#include "protoDefs.h"  // for ProtoSystemTime()
#include "protoTime.h"
#include <math.h>

/**
 * @class ProtoTimer
 *
 */
class ProtoTimer
{
    friend class ProtoTimerMgr;
    
    public:
        // Construction/destruction
        ProtoTimer();
        ~ProtoTimer();
        
        // NOTE: For VC++ Debug builds, you _cannot_ use pre-compiled
        // headers with this template code.  Also, "/ZI" or "/Z7" compile options 
        // must NOT be specified.  (or else VC++ experiences an "internal compiler error")
        
		// Listener Function 
		/**
		 * This function will define which method will be called upon a timer timing out
		 * Special notice should be taken of the boolean return value of this function.
		 * When returning true the number of repeats will be decremented and if below 0
		 * will deactivate the timer.  If in this function you are setting/changing intervals
		 * or repetions of the timer which invoked the call the function should return false
		 * to avoid standard exit actions.
		 */
		template <class listenerType>
        bool SetListener(listenerType* theListener, bool(listenerType::*timeoutHandler)(ProtoTimer&))
        {
            if (listener) delete listener;
            listener = theListener ? new LISTENER_TYPE<listenerType>(theListener, timeoutHandler) : NULL;
            return theListener ? (NULL != theListener) : true;
        }

        // Timer Parameters
        // Timer interval (seconds)
        
        /**
         * This method sets the interval (in seconds) after which the
         * timer will "fire" (i.e., the method specified by 
         * ProtoTimer::SetListener() is invoked).
         *
         * @param theInterval timer interval in seconds
         */ 
        void SetInterval(double theInterval) 
            {interval = (theInterval < 0.0) ? 0.0 : theInterval;}
         double GetInterval() const {return interval;}
        // Timer repetition (0 =  one shot, -1 = infinite repetition)
        void SetRepeat(int numRepeat) {repeat = numRepeat;}
        int GetRepeat() const {return repeat;}
        
        
        // These are provided for "advanced" timer monitoring/control
        void ResetRepeat() 
            {repeat_count = repeat;}
        void DecrementRepeatCount() 
            {repeat_count = (repeat_count > 0) ? repeat_count-- : repeat_count;}
        int GetRepeatCount() {return repeat_count;}
        // This must be used with wisdom!
        void SetRepeatCount(int repeatCount)
            {repeat_count = repeatCount;}
        
        //double GetTimeout() {return timeout;}
        
        // Activity status/control
        bool IsActive() const {return (NULL != mgr);}
        double GetTimeRemaining() const;
        bool Reschedule();
        void Scale(double factor);
        void Deactivate();
        
        // Ancillary
        void SetUserData(const void* userData) {user_data = userData;}
        const void* GetUserData() {return user_data;}
        
        // Installer commands
        enum Command {INSTALL, MODIFY, REMOVE};
            
    private:
        bool DoTimeout() {return listener ? listener->on_timeout(*this) : true;}
        class Listener
        {
            public:
                virtual ~Listener() {}
                virtual bool on_timeout(ProtoTimer& theTimer) = 0;
        };
        template <class listenerType>
        class LISTENER_TYPE : public Listener
        {
            public:
                LISTENER_TYPE(listenerType* theListener, bool(listenerType::*timeoutHandler)(ProtoTimer&))
                    : listener(theListener), timeout_handler(timeoutHandler) {}
                bool on_timeout(ProtoTimer& theTimer) 
                    {return (listener->*timeout_handler)(theTimer);}
                Listener* duplicate()
                    {return (static_cast<Listener*>(new LISTENER_TYPE<listenerType>(listener, timeout_handler)));}
            private:
                listenerType* listener;
                bool (listenerType::*timeout_handler)(ProtoTimer&);
        };
        Listener*                   listener;
        
        double                      interval;                
        int                         repeat;                  
        int                         repeat_count;            
        
        const void*                 user_data;
        ProtoTime                   timeout;                 
        bool                        is_precise;              
        class ProtoTimerMgr*        mgr;                     
        ProtoTimer*                 prev;                    
        ProtoTimer*                 next;                    
};  // end class ProtoTimer

/** 
 * @class ProtoTimerMgr
 */
class ProtoTimerMgr
{
    friend class ProtoTimer;
    
    public:
        ProtoTimerMgr();
        virtual ~ProtoTimerMgr();
        
        // ProtoTimer activation/deactivation
        virtual void ActivateTimer(ProtoTimer& theTimer);
        virtual void DeactivateTimer(ProtoTimer& theTimer);
        
        // Status
        double GetTimeRemaining() const 
            {return (short_head ? short_head->GetTimeRemaining() : -1.0);}
        
        // Call this when the timer mgr's one-shot system timer fires
        void OnSystemTimeout();
        
        // Special call to allow early dispatch of timeouts
        void DoSystemTimeout()
        {
            bool updateStatus = update_pending;
            update_pending = true;
            OnSystemTimeout();
            update_pending = updateStatus;
        }
        
        virtual void GetSystemTime(struct timeval& currentTime);
            
    protected:
        // System timer association/management definitions and routines
        virtual bool UpdateSystemTimer(ProtoTimer::Command command,
                                       double              delay) {return true;}
            
    private:
        // Methods used internally 
        void ReactivateTimer(ProtoTimer& theTimer, const ProtoTime& now);
        void InsertLongTimer(ProtoTimer& theTimer);
        void RemoveLongTimer(ProtoTimer& theTimer);
        void InsertShortTimer(ProtoTimer& theTimer);
        void RemoveShortTimer(ProtoTimer& theTimer);
        void Update();
        
        bool GetNextTimeout(ProtoTime& nextTimeout) const
        {
            if (NULL != short_head)
            {
                nextTimeout = short_head->timeout;
                return true;
            }
            else
            {
                return false;
            }
        }
        void GetPulseTime(ProtoTime& pulseTime) const
        {
            pulseTime = pulse_mark;
            pulseTime += (1.0 - pulse_timer.GetTimeRemaining());
        }
        bool OnPulseTimeout(ProtoTimer& theTimer);
        
        static const double PRECISION_TIME_THRESHOLD;
        
        
        void GetCurrentSystemTime(struct timeval& currentTime)
        {
#if  (defined(SIMULATE) && defined(NS2))
            GetSystemTime(currentTime);
#else
            ProtoSystemTime(currentTime);
#endif // if/else (SIMULATE && NS2)
        }
        
        // We need this to support some of our SIM code so it can
        // find its simulation context
        void GetCurrentProtoTime(ProtoTime& currentTime)
        {
#if  (defined(SIMULATE) && defined(NS2))
           GetSystemTime(currentTime.AccessTimeVal());
#else
           currentTime.GetCurrentTime();
#endif // if/else (SIMULATE && NS2)
        }
        
       
        // Member variables
        bool            update_pending;       
        bool            timeout_scheduled;                     
        ProtoTime       scheduled_timeout;                                   
        ProtoTimer      pulse_timer;  // one second pulse timer    
        ProtoTime       pulse_mark;                                
        ProtoTimer*     long_head;                                 
        ProtoTimer*     long_tail;                                 
        ProtoTimer*     short_head;                                
        ProtoTimer*     short_tail;  
};  // end class ProtoTimerMgr

#endif // _PROTO_TIMER
