#include "protoVif.h"
#ifdef UNIX
#include <unistd.h>      // for close()
#include <net/if.h>      // for ifreq
#include <fcntl.h>       // for open()
#include <sys/ioctl.h>   // for ioctl()
#include <stdlib.h>      // for system()
#include <stdio.h>       // for sprintf()
#endif

#ifdef LINUX
extern "C" 
{
#include <linux/if_tun.h>
}
#endif // LINUX

class UnixVif : public ProtoVif
{
  public:
    UnixVif();
    ~UnixVif();
            
    bool Open(char* vifName, unsigned int vifNameLenMax, ProtoAddress vifAddr, unsigned int maskLen);

    void Close();
    
    bool Write(const char* buffer, unsigned int numBytes);
    
    bool Read(char* buffer, unsigned int& numBytes);
          
}; // end class ProtoVif

ProtoVif* ProtoVif::Create()
{
    return static_cast<ProtoVif*>(new UnixVif);
}  // end ProtoVif::Create()

UnixVif::UnixVif()
{
}

UnixVif::~UnixVif()
{
}

bool UnixVif::Open(char* vifName, unsigned int vifNameLenMax, ProtoAddress vifAddr, unsigned int maskLen)
{
    
    
#ifdef LINUX    
    // 1) Open up the TUN/TAP device
    const char* devName = "/dev/net/tun";
    if ((descriptor = open(devName, O_RDWR)) < 0)
    {
        PLOG(PL_ERROR,"UnixVif::Open(%s) error: open(\"%s\") failed: %s\n", vifName, devName, GetErrorString());            
        return false;
    }

    // 2) Set up a TAP virtual interface with given "vifName"
    struct ifreq ifr;
    memset(&ifr, 0, sizeof(ifr));

    // Flags: IFF_TUN   - TUN device (no Ethernet headers) 
    //        IFF_TAP   - TAP device (includes ethernet headers)  
    //        IFF_NO_PI - Do not provide packet information  
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    strncpy(ifr.ifr_name, vifName,IFNAMSIZ);
    if (ioctl(descriptor, TUNSETIFF, &ifr) < 0)
    {
        PLOG(PL_ERROR, "UnixVif::Open(%s) error: ioctl(TUNSETIFF) failed: %s\n", vifName, GetErrorString());            
        Close(); 
        return false;
    }
#endif // LINUX
    
#ifdef MACOSX
    // On MacOSX, we have to iteratively try tap0, tap1, etc until we find one we can use
    if (vifNameLenMax < 7)
    {
        PLOG(PL_ERROR, "UnixVif::Open() error: vifName buffer too small!\n");
        return false;
    }
    char devName[256];
    for (int i = 0; i < 256; i++)
    {
        sprintf(devName, "/dev/tap%d", i);
        if ((descriptor = open(devName, O_RDWR)) < 0)
        {
            PLOG(PL_ERROR,"UnixVif::Open() error: open(\"%s\") failed: %s\n", devName, GetErrorString());            
            continue;
        }   
        sprintf((char*)vifName, "tap%d", i);
        break;
    }    
    if (INVALID_HANDLE == descriptor)
    {
        PLOG(PL_ERROR,"UnixVif::Open() error: no TAP device available!\n");            
        return false;
    }
#endif  // MACOSX  
    
    // 3) Configure the interface via "ifconfig" command
    // (TBD) Do this with an ioctl() call instead??
    char cmd[1024];
    sprintf(cmd, "/sbin/ifconfig %s %s/%d up", vifName, vifAddr.GetHostString(), maskLen);
    if (system(cmd) < 0)
    {
        PLOG(PL_ERROR, "UnixVif::Open(%s) error: \"%s\n\" failed: %s\n", vifName, cmd, GetErrorString());
        Close();
        return false;       
    }

    // 4) Make a call to "ProtoChannel::Open()" to install event dispatching if applicable
    if (!ProtoChannel::Open())
    {
        PLOG(PL_ERROR, "UnixVif::Open(%s) error: couldn't install ProtoChannel\n", vifName);
        Close();
        return false;    
    }
    else
    {
        return true;
    }
}  // end UnixVif::Open()

void UnixVif::Close()
{
    ProtoChannel::Close();
    close(descriptor);
    descriptor = INVALID_HANDLE;   
}  // end UnixVif::Close()

bool UnixVif::Write(const char* buffer, unsigned int numBytes) 
{
    int nWritten = write(descriptor, buffer, numBytes);
    if (nWritten != (int)numBytes)
    {
        PLOG(PL_ERROR,"UnixVif::Write() error: write() failure:%s\n", GetErrorString());
        return false;
    }    
    return true;
}  // end UnixVif::Write()

bool UnixVif::Read(char* buffer, unsigned int& numBytes)  
{
    int result = read(descriptor, buffer, numBytes);
    if (result < 0)
    {
        // (TBD) Automatically try again on errno == EINTR ???
        if (EAGAIN != errno)
            PLOG(PL_ERROR, "UnixVif::Read() error read() failure: %s\n", GetErrorString());
        return false;
    }
    numBytes = result;
    return true;
}  // end UnixVif::Read()
